# gemato: Profile support
# (c) 2017-2022 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import os.path


class DefaultProfile:
    """
    Profile is a class describing the specific properties of a directory
    tree. It is used when updating Manifests to determine the most
    correct behavior for a given use case.
    """

    name = 'default'

    def set_loader_options(self, loader):
        """
        Alter loader @loader with profile-specific options. This
        is called after applying the user-specified options, so it
        should check them for None if overwriting is not desired.
        """
        pass

    def get_entry_type_for_path(self, path):
        """
        Get Manifest entry type appropriate for the specified path.
        Must return an appropriate Manifest tag for file-style entry
        (i.e. one of DATA, MISC, EBUILD, AUX).
        """
        return 'DATA'

    def want_manifest_in_directory(self, relpath, dirnames, filenames):
        """
        Determine whether a Manifest file is expected in the specified
        directory. @relpath is the relative path to the directory,
        @dirnames and @filenames list respectively all directories
        and files directly underneath it.

        Should return True if Manifest is expected, False otherwise.
        If True is returned and the directory does not contain a single
        file of Manifest type, a new one will be created as 'Manifest'.
        """
        return False

    def get_ignore_paths_for_new_manifest(self, relpath):
        """
        Get the list of IGNORE paths that should be added to the newly
        created Manifest in directory @relpath. The paths must be
        relative to @relpath.

        This function is only called when a new Manifest file is being
        created.
        """
        return ()

    def want_compressed_manifest(self, relpath, manifest, unc_size,
                                 compress_watermark):
        """
        Determine whether the specified Manifest (at @relpath) can
        be compressed. @manifest is the Manifest instance. @unc_size
        specified the uncompressed data size, and @compress_watermark
        is the watermark value at the time of invocation.

        Should return True to compress Manifest, False to uncompress it
        or None to leave as-is.
        """
        # Compress files above the watermark but not the top-level
        # Manifest. We only check for basename -- if it was compressed
        # already, we do not change that.
        return (unc_size >= compress_watermark and relpath != 'Manifest')


class EbuildRepositoryProfile(DefaultProfile):
    """
    A profile suited for a modern ebuild repository.
    """

    name = 'ebuild'

    def want_manifest_in_directory(self, relpath, dirnames, filenames):
        # a quick way to catch most of packages and ::gentoo categories
        if 'metadata.xml' in filenames:
            return True
        spl = relpath.split(os.path.sep)
        # top level directories...
        if len(spl) == 1:
            # with any subdirectories (categories, metadata, profiles)
            if len(dirnames) > 0:
                return True
            # plus some unconditional standard directories
            if relpath in ('eclass', 'licenses', 'metadata',
                           'profiles'):
                return True
        elif len(spl) == 2:
            # 'slow' way of detecting package directories
            if any(f.endswith('.ebuild') for f in filenames):
                return True
            # some standard directories worth separate Manifests
            if spl[0] == 'metadata' and spl[1] in ('dtd',
                                                   'glsa',
                                                   'md5-cache',
                                                   'news',
                                                   'xml-schema'):
                return True
        elif len(spl) == 3:
            # metadata cache -> per-directory Manifests
            if spl[0:2] == ['metadata', 'md5-cache']:
                return True
        return False

    def get_ignore_paths_for_new_manifest(self, relpath):
        if relpath == '':
            # traditionally present in /usr/portage
            return ('distfiles', 'local', 'lost+found', 'packages')
        elif relpath == 'metadata':
            return ('timestamp', 'timestamp.chk', 'timestamp.commit',
                    'timestamp.x')
        elif relpath in ('metadata/dtd', 'metadata/glsa',
                         'metadata/news', 'metadata/xml-schema'):
            return ('timestamp.chk', 'timestamp.commit')
        return ()

    def set_loader_options(self, loader):
        if loader.hashes is None:
            # layout.conf as of 2017-11-21
            loader.hashes = ['BLAKE2B', 'SHA512']
        if loader.sort is None:
            loader.sort = True
        if loader.compress_watermark is None:
            # 128 should be a safe value where gzip can actually
            # gain anything without making things worse by overhead
            loader.compress_watermark = 128
        if loader.compress_format is None:
            loader.compress_format = 'gz'


class BackwardsCompatEbuildRepositoryProfile(EbuildRepositoryProfile):
    """
    A profile for ebuild repository that maintains compatibility
    with Manifest2 format.
    """

    name = 'old-ebuild'

    def get_entry_type_for_path(self, path):
        spl = path.split(os.path.sep)
        if len(spl) == 3:
            if path.endswith('.ebuild'):
                return 'EBUILD'
            elif spl[2] == 'metadata.xml':
                return 'MISC'
        if spl[2:3] == ['files']:
            return 'AUX'

        return (super().get_entry_type_for_path(path))

    def want_compressed_manifest(self, relpath, manifest, unc_size,
                                 compress_watermark):
        for e in manifest.entries:
            # disable compression in package directories
            if e.tag == 'EBUILD':
                return False

        return (
            super().want_compressed_manifest(relpath,
                                             manifest,
                                             unc_size,
                                             compress_watermark))


PROFILE_MAPPING = {
    getattr(profile, 'name'): profile
    for profile in (DefaultProfile,
                    EbuildRepositoryProfile,
                    BackwardsCompatEbuildRepositoryProfile,
                    )}


def get_profile_by_name(name):
    return PROFILE_MAPPING[name]()



#!/usr/bin/env python

import os.path
import sys
import timeit

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

import gemato.recursiveloader


def benchmark(path):
    m = gemato.recursiveloader.ManifestRecursiveLoader(path)
    print(f'load-dict: {timeit.timeit(m.get_file_entry_dict, number=1)}')
    print(f'verify: {timeit.timeit(m.assert_directory_verifies, number=1)}')


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print(f'Usage: {sys.argv[0]} <top-level-manifest>')
        sys.exit(1)

    benchmark(sys.argv[1])



#!/usr/bin/env python

import os.path
import sys
import timeit

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

import gemato.hash


def benchmark_one(path, hashes):
	f = lambda: gemato.hash.hash_path(path, hashes)
	print(f"{hashes} -> [ ", end='', flush=True)
	results = []
	for t in (timeit.timeit(f, number=1) for i in range(5)):
		print(f"{t:.4}", end=" ", flush=True)
		results.append(t)
	print(f"] -> min: {min(results):.4}")


def benchmark(path, hash_sets):
	if not hash_sets:
		hash_sets = [
			["sha256"],
			["sha256", "blake2b"],
			["sha256", "blake2b", "sha512"],
		]
	else:
		hash_sets = [x.split() for x in hash_sets]

	for hashes in hash_sets:
		benchmark_one(path, hashes)


if __name__ == '__main__':
	if len(sys.argv) < 2:
		print(f'Usage: {sys.argv[0]} <test-file> [<hash-set>...]')
		sys.exit(1)

	benchmark(sys.argv[1], sys.argv[2:])



# gemato: Test utility functions
# (c) 2017-2023 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import errno
import functools
import os
import os.path
import random
import stat
import threading

import pytest

from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import urlparse, parse_qs


def disallow_writes(path):
    """Mark path non-writable, recursively"""
    for dirpath, dirs, files in os.walk(path, topdown=False):
        for f in files + dirs:
            st = os.lstat(os.path.join(dirpath, f))
            if not stat.S_ISLNK(st.st_mode):
                os.chmod(os.path.join(dirpath, f),
                         st.st_mode & ~0o222)
    os.chmod(path, 0o555)


class HKPServerRequestHandler(BaseHTTPRequestHandler):
    def __init__(self, keys, *args, **kwargs):
        self.keys = keys
        BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def log_message(self, *args, **kwargs):
        pass

    def do_GET(self):
        try:
            parsed = urlparse(self.path)
            assert parsed.path == '/pks/lookup'

            qs = parse_qs(parsed.query)
            assert qs.get('op') == ['get']
            assert len(qs.get('search', [])) == 1

            key = qs['search'][0]
            assert key.startswith('0x')
            key = key[2:]
        except AssertionError:
            self.send_error(400, "Bad request")
            return

        if key not in self.keys:
            self.send_error(404, "Not found")
            return

        self.send_response(200, "OK")
        self.send_header("Content-type", "application/pgp-keys")
        self.end_headers()
        # note: technically we should be using ASCII armor here
        # but GnuPG seems happy with the binary form too
        self.wfile.write(self.keys[key])
        self.wfile.flush()


class HKPServer:
    def __init__(self):
        self.keys = {}
        self.addr = None

    def start(self):
        # try 10 randomly selected ports before giving up
        for port in random.sample(range(1024, 32768), 10):
            try:
                self.server = HTTPServer(
                    ('127.0.0.1', port),
                    functools.partial(HKPServerRequestHandler, self.keys))
            except OSError as e:
                if e.errno != errno.EADDRINUSE:
                    pytest.skip(
                        f'Unable to bind the HKP server: {e}')
            else:
                break
        else:
            pytest.skip('Unable to find a free port for HKP server')

        self.addr = f'hkp://127.0.0.1:{port}'
        self.thread = threading.Thread(target=self.server.serve_forever)
        self.thread.start()

    def stop(self):
        assert self.addr is not None
        self.server.shutdown()
        self.server.server_close()
        self.thread.join()
        self.addr = None



# gemato: CLI routines
# (c) 2017-2023 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later


import argparse
import datetime
import logging
import multiprocessing
import os.path
import signal
import subprocess
import sys
import timeit

from pathlib import Path

from gemato.exceptions import GematoException, ManifestMismatch
from gemato.find_top_level import find_top_level_manifest
from gemato.hash import hash_file, hash_path
from gemato.manifest import (
    ManifestFileEntry,
    manifest_hashes_to_hashlib,
    )
from gemato.openpgp import (
    OpenPGPEnvironment,
    OpenPGPSystemEnvironment,
    IsolatedGPGEnvironment,
    )
from gemato.profile import get_profile_by_name
from gemato.recursiveloader import ManifestRecursiveLoader


LOGGER = logging.getLogger(__name__)


def verify_failure(e):
    LOGGER.error(e)
    return False


class GematoCommand:
    """
    Base class for commands supported by gemato.
    """

    @property
    def name(self):
        """
        Command name. Used on the command-line
        """
        pass

    @property
    def help(self):
        """
        Command description for --help.
        """
        pass

    def add_options(self, subp):
        """
        Add options specific to the command to subparser @subp.
        """
        subp.add_argument("--debug", action="store_const",
                          dest="log_level", const=logging.DEBUG,
                          default=logging.INFO,
                          help='Enable debugging output')
        subp.add_argument("-q", "--quiet", action="store_const",
                          dest="log_level", const=logging.WARNING,
                          help="Only emit warning or error messages")

    def parse_args(self, args, argp):
        """
        Process command-line arguments @args. @argp is the argparse
        instance provided for error reporting.
        """
        # allow an easy override for users when gemato is called
        # via scripts
        if os.environ.get("GEMATO_DEBUG"):
            args.log_level = logging.DEBUG
        logging.getLogger().setLevel(args.log_level)

    def __call__(self):
        """
        Perform the command. Returns the exit status.
        """
        pass

    def cleanup(self):
        """
        Perform any cleanups necessary. Called on program termination.
        """
        pass


class BaseOpenPGPMixin:
    """
    A base mixin that adds logic to load and use OpenPGP keys.
    """

    def __init__(self):
        super().__init__()
        self.openpgp_env = None

    def add_options(self, subp):
        super().add_options(subp)

        default_timeout = 180

        subp.add_argument(
            '-K', '--openpgp-key',
            help='Use only the OpenPGP key(s) from a specific file')
        subp.add_argument(
            '--proxy',
            help='Use HTTP proxy')
        subp.add_argument(
            "--timeout",
            default=default_timeout,
            type=float,
            help="Connection timeout (for WKD requests, in seconds, "
                 f"default: {default_timeout})")

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        # use isolated environment if key is specified;
        # system environment otherwise
        if args.openpgp_key is not None:
            env_class = OpenPGPEnvironment
        else:
            env_class = OpenPGPSystemEnvironment
        self.openpgp_env = env_class(debug=args.log_level == logging.DEBUG,
                                     proxy=args.proxy,
                                     timeout=args.timeout)

        if args.openpgp_key is not None:
            with open(args.openpgp_key, 'rb') as f:
                self.openpgp_env.import_key(f)

    def cleanup(self):
        super().cleanup()

        if self.openpgp_env is not None:
            self.openpgp_env.close()


class VerifyingOpenPGPMixin(BaseOpenPGPMixin):
    """
    Verification-class OpenPGP mixin. Additionally refreshes keys.
    """

    def add_options(self, subp):
        super().add_options(subp)

        subp.add_argument(
            '-R', '--no-refresh-keys', action='store_false',
            dest='refresh_keys',
            help='Disable refreshing OpenPGP key (prevents network '
                 'access, applicable when using -K only)')
        subp.add_argument(
            '-W', '--no-wkd', action='store_false',
            dest='allow_wkd',
            help='Do not attempt to use WKD to refetch keys (use '
                 'keyservers only)')
        subp.add_argument(
            '--keyserver',
            help='Force custom keyserver URL')

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        if args.openpgp_key is not None:
            # always refresh keys to check for revocation
            # (unless user specifically asked us not to)
            if args.refresh_keys:
                LOGGER.info('Refreshing keys...')
                self.openpgp_env.refresh_keys(allow_wkd=args.allow_wkd,
                                              keyserver=args.keyserver)
                LOGGER.info('Keys refreshed.')

    def print_signatures(self, sigs):
        for i, sig in enumerate(sigs):
            if len(sigs) > 1:
                LOGGER.info(f"-- signature {i}")
            LOGGER.info(f"- status: {sig.sig_status}")
            LOGGER.info(f"- valid: {sig.valid_sig}, "
                        f"trusted: {sig.trusted_sig}")
            if sig.valid_sig:
                LOGGER.info("- primary key: "
                            f"{sig.primary_key_fingerprint}")
                LOGGER.info(f"- subkey: {sig.fingerprint}")
                LOGGER.info(f"- timestamp: {sig.timestamp} UTC")
            if sig.key_expiration is not None:
                LOGGER.info(f"- key expiration: {sig.key_expiration} UTC")


class BaseManifestLoaderMixin:
    """
    Mixin for commands using RecursiveManifestLoader class.
    """

    def add_options(self, subp):
        super().add_options(subp)

        subp.add_argument(
            '-j', '--jobs', type=int,
            help=f'Specify the maximum number of parallel jobs to use '
                 f'(default: {multiprocessing.cpu_count()})')
        subp.add_argument(
            '-x', '--one-file-system', action='store_true',
            help='Do not cross filesystem boundaries (report an error '
                 'instead)')
        secugroup = subp.add_mutually_exclusive_group()
        secugroup.add_argument(
            "--require-secure-hashes", action="store_true",
            default=None,
            help="Require using secure hashes (default if Manifest is signed)")
        secugroup.add_argument(
            "--no-require-secure-hashes", action="store_false",
            dest="require_secure_hashes",
            help="Do not require using secure hashes (default if Manifest "
                 "is not signed)")

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        self.init_kwargs = {}
        if args.jobs is not None:
            if args.jobs < 1:
                argp.error('--jobs must be positive')
            self.init_kwargs['max_jobs'] = args.jobs
        if args.one_file_system:
            self.init_kwargs['allow_xdev'] = False
        if args.require_secure_hashes is not None:
            self.init_kwargs["require_secure_hashes"] = (
                args.require_secure_hashes)


class VerifyCommand(BaseManifestLoaderMixin, VerifyingOpenPGPMixin,
                    GematoCommand):
    name = 'verify'
    help = 'Verify one or more directories against Manifests'

    def add_options(self, verify):
        super().add_options(verify)

        verify.add_argument(
            'paths', nargs='*', default=['.'],
            help='Paths to verify (defaults to "." if none specified)')
        verify.add_argument(
            '-k', '--keep-going', action='store_true',
            help='Continue reporting errors rather than terminating '
                 'on the first failure')
        verify.add_argument(
            '-P', '--no-openpgp-verify', action='store_false',
            dest='openpgp_verify',
            help='Disable OpenPGP verification of signed Manifests')
        verify.add_argument(
            '-s', '--require-signed-manifest', action='store_true',
            help='Require that the top-level Manifest is OpenPGP signed')

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        self.paths = args.paths
        self.require_signed_manifest = args.require_signed_manifest
        self.kwargs = {}
        self.init_kwargs['openpgp_env'] = self.openpgp_env

        if args.keep_going:
            self.kwargs['fail_handler'] = verify_failure
        if not args.openpgp_verify:
            self.init_kwargs['verify_openpgp'] = False

    def __call__(self):
        super().__call__()

        ret = True

        for p in self.paths:
            tlm = find_top_level_manifest(p)
            if tlm is None:
                LOGGER.error(f'Top-level Manifest not found in {p}')
                return 1

            LOGGER.info(f"Using top-level Manifest: {tlm}")

            start = timeit.default_timer()
            m = ManifestRecursiveLoader(tlm, **self.init_kwargs)
            if self.require_signed_manifest and not m.openpgp_signed:
                LOGGER.error(f'Top-level Manifest {tlm} is not '
                             f'OpenPGP signed')
                return 1

            ts = m.find_timestamp()
            if ts:
                LOGGER.info(f'Manifest timestamp: {ts.ts} UTC')

            if m.openpgp_signed:
                LOGGER.info('Valid OpenPGP signature found:')
                self.print_signatures(m.openpgp_signature)

            LOGGER.info(f'Verifying {p}...')

            tlmdir = os.path.dirname(tlm)
            relpath = os.path.relpath(p, tlmdir)
            if relpath == '.':
                relpath = ''
            try:
                ret &= m.assert_directory_verifies(relpath, **self.kwargs)
            except ManifestMismatch as e:
                apparent_path = os.path.join(tlmdir, e.path)
                real_path = os.path.realpath(apparent_path)
                if apparent_path != real_path:
                    LOGGER.warning(
                        f"Path contains symlinks. Real file path: {real_path}")
                raise

            stop = timeit.default_timer()
            LOGGER.info(f'{p} verified in {stop - start:.2f} seconds')

        return 0 if ret else 1


class BaseUpdateMixin(BaseManifestLoaderMixin, BaseOpenPGPMixin):
    """
    A mixin that adds common bits for update-class commands.
    """

    def add_options(self, update):
        super().add_options(update)

        update.add_argument(
            '-c', '--compress-watermark', type=int,
            help='Minimum Manifest size for files to be compressed')
        update.add_argument(
            '-C', '--compress-format',
            help='Format for compressed files (e.g. "gz", "bz2"...)')
        update.add_argument(
            '-f', '--force-rewrite', action='store_true',
            help='Force rewriting all the Manifests, even if they did '
                 'not change')
        update.add_argument(
            '-H', '--hashes',
            help='Whitespace-separated list of hashes to use')
        update.add_argument(
            '-k', '--openpgp-id',
            help='Use the specified OpenPGP key (by ID or user)')
        update.add_argument(
            '-p', '--profile',
            help='Use the specified profile ("default", "ebuild", '
                 '"old-ebuild"...)')
        signgroup = update.add_mutually_exclusive_group()
        signgroup.add_argument(
            '-s', '--sign', action='store_true',
            default=None,
            help='Force signing the top-level Manifest')
        signgroup.add_argument(
            '-S', '--no-sign', action='store_false',
            dest='sign',
            help='Disable signing the top-level Manifest')
        update.add_argument(
            '-t', '--timestamp', action='store_true',
            help='Include TIMESTAMP entry in Manifest')

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        self.timestamp = args.timestamp

        self.save_kwargs = {}
        self.init_kwargs['openpgp_env'] = self.openpgp_env

        if args.hashes is not None:
            self.init_kwargs['hashes'] = args.hashes.split()
        if args.compress_watermark is not None:
            if args.compress_watermark < 0:
                argp.error('--compress-watermark must not be negative!')
            self.init_kwargs['compress_watermark'] = args.compress_watermark
        if args.compress_format is not None:
            self.init_kwargs['compress_format'] = args.compress_format
        if args.force_rewrite:
            self.save_kwargs['force'] = True
        if args.openpgp_id is not None:
            self.init_kwargs['openpgp_keyid'] = args.openpgp_id
        if args.profile is not None:
            self.init_kwargs['profile'] = (
                get_profile_by_name(args.profile))
        if args.sign is not None:
            self.init_kwargs['sign_openpgp'] = args.sign


class UpdateCommand(BaseUpdateMixin, GematoCommand):
    name = 'update'
    help = 'Update the Manifest entries for one or more directory trees'

    def add_options(self, update):
        super().add_options(update)

        update.add_argument(
            'paths', nargs='*', default=['.'],
            help='Paths to update (defaults to "." if none specified)')
        update.add_argument(
            '-i', '--incremental', action='store_true',
            help='Perform incremental update by comparing mtimes '
                 'against TIMESTAMP')

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        self.paths = args.paths
        self.incremental = args.incremental

    def __call__(self):
        super().__call__()

        for p in self.paths:
            tlm = find_top_level_manifest(p)
            if tlm is None:
                LOGGER.error(f'Top-level Manifest not found in {p}')
                return 1

            start = timeit.default_timer()
            m = ManifestRecursiveLoader(tlm, **self.init_kwargs)

            # if not specified by user, profile must set it
            if m.hashes is None:
                LOGGER.error('--hashes must be specified if not '
                             'implied by --profile')
                return 1

            relpath = os.path.relpath(p, os.path.dirname(tlm))
            if relpath == '.':
                relpath = ''
            if self.timestamp and relpath != '':
                LOGGER.error('Timestamp can only be updated if doing '
                             'full-tree update')
                return 1

            update_kwargs = {}
            if self.incremental:
                if relpath != '':
                    LOGGER.error('Incremental works only for '
                                 'full-tree update')
                    return 1
                last_ts = m.find_timestamp()
                if last_ts is None:
                    LOGGER.error('Incremental specified but no '
                                 'timestamp in Manifest')
                    return 1
                update_kwargs['last_mtime'] = last_ts.ts.timestamp()

            LOGGER.info(f'Updating Manifests in {p}...')

            start_ts = datetime.datetime.now(datetime.timezone.utc)
            m.update_entries_for_directory(relpath, **update_kwargs)

            # write TIMESTAMP if requested, or if already there
            if relpath != '':
                # skip timestamp if not doing full update
                pass
            elif self.timestamp:
                m.set_timestamp(start_ts)
            else:
                ts = m.find_timestamp()
                if ts is not None:
                    ts.ts = start_ts

            m.save_manifests(**self.save_kwargs)

            stop = timeit.default_timer()
            LOGGER.info(f'{p} updated in {stop - start:.2f} seconds')

        return 0


class CreateCommand(BaseUpdateMixin, GematoCommand):
    name = 'create'
    help = 'Create a Manifest tree starting at the specified file'

    def add_options(self, create):
        super().add_options(create)

        create.add_argument(
            'paths', nargs='*', default=['.'],
            help='Paths to create Manifest in (defaults to "." '
                 'if none specified)')

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        self.init_kwargs['allow_create'] = True
        self.paths = args.paths

    def __call__(self):
        super().__call__()

        for p in self.paths:
            start = timeit.default_timer()
            m = ManifestRecursiveLoader(
                os.path.join(p, 'Manifest'), **self.init_kwargs)

            # if not specified by user, profile must set it
            if m.hashes is None:
                LOGGER.error('--hashes must be specified if not '
                             'implied by --profile')
                return 1

            LOGGER.info(f'Creating Manifests in {p}...')

            start_ts = datetime.datetime.now(datetime.timezone.utc)
            m.update_entries_for_directory()

            # write TIMESTAMP if requested
            if self.timestamp:
                m.set_timestamp(start_ts)

            m.save_manifests(**self.save_kwargs)

            stop = timeit.default_timer()
            LOGGER.info(f'{p} updated in {stop - start:.2f} seconds')

        return 0


class HashCommand(GematoCommand):
    name = 'hash'
    help = 'Generate hashes for specified file(s) and/or stdin'

    def add_options(self, subp):
        super().add_options(subp)

        subp.add_argument(
            'paths', nargs='*', default=['-'],
            help='Paths to hash (defaults to "-" (stdin) if not '
                 'specified)')
        subp.add_argument(
            '-H', '--hashes', required=True,
            help='Whitespace-separated list of hashes to use')

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        self.hashes = sorted(args.hashes.split())
        self.paths = args.paths

    def __call__(self):
        super().__call__()

        hashlib_hashes = list(manifest_hashes_to_hashlib(self.hashes))
        hashlib_hashes.append('__size__')

        for p in self.paths:
            if p == '-':
                h = hash_file(sys.stdin.buffer, hashlib_hashes)
            else:
                h = hash_path(p, hashlib_hashes)

            sz = h.pop('__size__')
            e = ManifestFileEntry(
                p, sz,
                {mh: h[hh] for mh, hh in zip(self.hashes, hashlib_hashes)})
            print(' '.join(e.to_list('DATA' if p != '-' else 'STDIN')))


class OpenPGPVerifyCommand(VerifyingOpenPGPMixin, GematoCommand):
    name = 'openpgp-verify'
    help = ('Verify OpenPGP signatures embedded in specified file(s) '
            'and/or stdin')

    def add_options(self, subp):
        super().add_options(subp)

        subp.add_argument(
            "--no-require-all-good",
            dest="require_all_good",
            action="store_false",
            default=True,
            help="Require only one good signature on multi-signature files")
        subp.add_argument(
            'paths', nargs='*', default=['-'],
            help='Paths to hash (defaults to "-" (stdin) if not '
                 'specified)')

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        self.paths = args.paths
        self.require_all_good = args.require_all_good

    def __call__(self):
        super().__call__()

        ret = True

        for p in self.paths:
            if p == '-':
                f = sys.stdin
            else:
                f = open(p)

            try:
                try:
                    sigs = self.openpgp_env.verify_file(
                        f, require_all_good=self.require_all_good)
                except GematoException as e:
                    LOGGER.error(
                        f'OpenPGP verification failed for {p}:\n{e}')
                    ret = False
                else:
                    LOGGER.info(
                        f'Valid OpenPGP signature found in {p}:')
                    self.print_signatures(sigs)
            finally:
                if p != '-':
                    f.close()

        return 0 if ret else 1


class OpenPGPVerifyDetachedCommand(VerifyingOpenPGPMixin, GematoCommand):
    name = "openpgp-verify-detached"
    help = "Verify a file against detached OpenPGP signature "

    def add_options(self, subp):
        super().add_options(subp)

        subp.add_argument(
            "--no-require-all-good",
            dest="require_all_good",
            action="store_false",
            default=True,
            help="Require only one good signature on multi-signature files")
        subp.add_argument(
            "signature_file",
            type=Path,
            help="Path to the file containing the OpenPGP signature")
        subp.add_argument(
            "data_file",
            type=argparse.FileType("rb"),
            help="Path to the file to verify or \"-\" for stdin")

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        self.signature_file = args.signature_file
        self.data_file = args.data_file
        self.require_all_good = args.require_all_good

    def __call__(self):
        super().__call__()

        try:
            sigs = self.openpgp_env.verify_detached(
                self.signature_file, self.data_file,
                require_all_good=self.require_all_good)
        except GematoException as e:
            LOGGER.error(
                f"OpenPGP verification failed for {self.data_file} "
                f"(sig in {self.signature_file}):\n{e}")
            return 1
        else:
            LOGGER.info(
                f"File {self.data_file.name} verified successfully against "
                f"the signature in {self.signature_file}:")
            self.print_signatures(sigs)

        return 0


class GnuPGWrapCommand(VerifyingOpenPGPMixin, GematoCommand):
    name = 'gpg-wrap'
    help = ('Run specified command with GNUPGHOME set to OpenPGP '
            'verification environment (with keys loaded)')

    def add_options(self, subp):
        super().add_options(subp)

        subp.add_argument(
            'argv', nargs='+',
            help='Command to execute')

    def parse_args(self, args, argp):
        super().parse_args(args, argp)

        if args.openpgp_key is None:
            argp.error('gpg-wrap requires --openpgp-key to be specified')
        self.argv = args.argv

    def __call__(self):
        super().__call__()
        assert isinstance(self.openpgp_env, IsolatedGPGEnvironment)
        os.environ['GNUPGHOME'] = self.openpgp_env._home
        p = subprocess.Popen(self.argv)
        ret = p.wait()
        if ret < 0:
            if hasattr(signal, 'strsignal'):
                sig = signal.strsignal(-ret)
            else:
                sig = -ret
            LOGGER.error(
                f'Child process terminated due to signal: {sig}')
        return ret


def main(argv):
    argp = argparse.ArgumentParser(
            prog=argv[0],
            description='Gentoo Manifest Tool')
    subp = argp.add_subparsers()

    commands = [VerifyCommand,
                UpdateCommand,
                CreateCommand,
                HashCommand,
                OpenPGPVerifyCommand,
                OpenPGPVerifyDetachedCommand,
                GnuPGWrapCommand,
                ]
    for cmdclass in commands:
        cmd = cmdclass()
        cmdp = subp.add_parser(cmd.name, help=cmd.help)
        cmd.add_options(cmdp)
        cmdp.set_defaults(cmd=cmd)

    vals = argp.parse_args(argv[1:])
    if not hasattr(vals, 'cmd'):
        argp.error('No function specified')
    try:
        assert isinstance(vals.cmd, GematoCommand)

        try:
            vals.cmd.parse_args(vals, argp)
            return vals.cmd()
        finally:
            vals.cmd.cleanup()
    except GematoException as e:
        LOGGER.error(e)
        return 1


def setuptools_main():
    try:
        from rich.logging import RichHandler
    except ImportError:
        format_args = {
            "format": "[{levelname:>7}] {message}",
            "style": "{",
        }
    else:
        format_args = {
            "format": "{message}",
            "style": "{",
            "handlers": [RichHandler(show_time=False, show_path=False)],
        }

    logging.basicConfig(
        level=logging.INFO,
        **format_args)
    sys.exit(main(sys.argv))


if __name__ == '__main__':
    setuptools_main()



#!/usr/bin/env python

import glob
import os
import os.path
import sys

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

import gemato.hash


def write_manifest_entry(manifest_file, t, path, relpath, hashes):
    checksums = gemato.hash.hash_path(path,
            [x.lower() for x in hashes] + ['__size__'])
    hashvals = []
    for h in hashes:
        hashvals += [h, checksums[h.lower()]]
    manifest_file.write('{} {} {} {}\n'.format(t, relpath,
        checksums['__size__'], ' '.join(hashvals)))


def write_manifest_entries_for_dir(manifest_file, topdir, hashes):
    for dirpath, dirs, files in os.walk(topdir):
        if dirpath != topdir:
            for f in files:
                if f.startswith('Manifest'):
                    fp = os.path.join(dirpath, f)
                    write_manifest_entry(manifest_file, 'MANIFEST',
                            fp, os.path.relpath(fp, topdir), hashes)
                    # do not descend
                    dirs.clear()
                    skip = True
                    break
            else:
                skip = False
            if skip:
                continue

        for f in files:
            if f.startswith('Manifest'):
                continue
            fp = os.path.join(dirpath, f)
            write_manifest_entry(manifest_file, 'DATA',
                    fp, os.path.relpath(fp, topdir), hashes)


def gen_metamanifests(top_dir, hashes):
    with open(os.path.join(top_dir, 'profiles/categories')) as f:
        categories = [x.strip() for x in f]

    alldirs = []

    # we assume every package has thick Manifests already, so we just
    # need to Manifest the Manifests
    for c in categories:
        alldirs.append(c)
        alldirs.append(os.path.join('metadata/md5-cache', c))

    # Manifest a few big dirs separately
    alldirs.extend(['eclass', 'licenses', 'metadata/md5-cache', 'metadata/glsa',
            'metadata/news', 'metadata', 'profiles'])

    for bm in alldirs:
        bmdir = os.path.join(top_dir, bm)
        if not list(glob.glob(os.path.join(bmdir, 'Manifest*'))):
            with open(os.path.join(bmdir, 'Manifest'), 'w') as f:
                write_manifest_entries_for_dir(f, bmdir, hashes)

    with open(os.path.join(top_dir, 'Manifest'), 'w') as f:
        write_manifest_entries_for_dir(f, top_dir, hashes)


if __name__ == '__main__':
    if len(sys.argv) != 3:
        print(f'Usage: {sys.argv[0]} <rsync-path> <hashes>')
        sys.exit(1)

    gen_metamanifests(sys.argv[1], sys.argv[2].split())



# gemato: Top-level Manifest finding tests
# (c) 2017-2022 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import gzip
import os
import os.path

import pytest

from gemato.find_top_level import find_top_level_manifest

from tests.testutil import disallow_writes


@pytest.fixture(scope='module')
def plain_tree(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp('find-top-level-plain-')
    for d in ('empty-subdir',
              'manifest-subdir',
              'deep/manifest-subdir',
              'ignored-dir',
              'ignored-dir/subdir',
              'ignored-dir-not',
              'ignored-empty-dir'):
        os.makedirs(tmp_path / d)
    with open(tmp_path / 'Manifest', 'w') as f:
        f.write('''
IGNORE ignored-dir
IGNORE ignored-empty-dir
''')
    for f in ('manifest-subdir/Manifest',
              'deep/manifest-subdir/Manifest',
              'ignored-dir/Manifest'):
        with open(tmp_path / f, 'w'):
            pass
    disallow_writes(tmp_path)
    yield tmp_path


@pytest.mark.parametrize(
    'start_dir,expected',
    [('.', 'Manifest'),
     ('empty-subdir', 'Manifest'),
     ('manifest-subdir', 'Manifest'),
     ('deep/manifest-subdir', 'Manifest'),
     ('ignored-dir', 'ignored-dir/Manifest'),
     ('ignored-dir/subdir', 'ignored-dir/Manifest'),
     ('ignored-dir-not', 'Manifest'),
     ('ignored-empty-dir', None),
     ])
def test_find_top_level_manifest(plain_tree, start_dir, expected):
    """Test finding top-level Manifest from plain directory tree"""
    mpath = find_top_level_manifest(plain_tree / start_dir)
    if mpath is not None:
        mpath = os.path.relpath(mpath, plain_tree)
    assert mpath == expected


def test_unreadable_manifest(tmp_path):
    """Test failure when one of Manifest files is not readable"""
    with open(tmp_path / 'Manifest', 'w') as f:
        os.fchmod(f.fileno(), 0)
    with pytest.raises(PermissionError):
        find_top_level_manifest(tmp_path)


def test_empty_tree(tmp_path):
    """Test working on empty tree without a Manifest file"""
    assert find_top_level_manifest(tmp_path) is None


def test_root_directory(tmp_path):
    """Test that things do not explode when running on /"""
    if os.path.exists('/Manifest'):
        pytest.skip('Manifest is present in system root ("/")')
    assert find_top_level_manifest('/') is None


def test_cross_device(tmp_path):
    """Test that device boundaries are not crossed"""
    if not os.path.ismount('/proc'):
        pytest.skip('/proc is not a mount point')
    with open(tmp_path / 'Manifest', 'w'):
        pass
    os.symlink('/proc', tmp_path / 'test')
    assert find_top_level_manifest(tmp_path / 'test') is None


@pytest.fixture(scope='module')
def compressed_manifest_tree(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp('find-top-level-compressed-')
    for d in ('empty-subdir',
              'manifest-subdir',
              'deep/manifest-subdir',
              'ignored-dir',
              'ignored-dir/subdir',
              'ignored-dir-not',
              'ignored-empty-dir'):
        os.makedirs(tmp_path / d)
    with gzip.GzipFile(tmp_path / 'Manifest.gz', 'wb') as f:
        f.write(b'''
IGNORE ignored-dir
IGNORE ignored-empty-dir
''')
    with open(tmp_path / 'manifest-subdir/Manifest', 'wb'):
        pass
    for f in ('deep/manifest-subdir/Manifest.gz',
              'ignored-dir/Manifest.gz'):
        with gzip.GzipFile(tmp_path / f, 'w'):
            pass
    disallow_writes(tmp_path)
    yield tmp_path


@pytest.mark.parametrize(
    'start_dir,allow_compressed,expected',
    [('.', False, None),
     ('.', True, 'Manifest.gz'),
     ('empty-subdir', True, 'Manifest.gz'),
     ('manifest-subdir', True, 'Manifest.gz'),
     ('deep/manifest-subdir', True, 'Manifest.gz'),
     ('ignored-dir', True, 'ignored-dir/Manifest.gz'),
     ('ignored-dir/subdir', True, 'ignored-dir/Manifest.gz'),
     ('ignored-dir-not', True, 'Manifest.gz'),
     ('ignored-empty-dir', True, None),
     ])
def test_find_compressed_top_level_manifest(compressed_manifest_tree,
                                            start_dir,
                                            allow_compressed,
                                            expected):
    """Test finding compressed top-level Manifest """
    mpath = find_top_level_manifest(compressed_manifest_tree / start_dir,
                                    allow_compressed=allow_compressed)
    if mpath is not None:
        mpath = os.path.relpath(mpath, compressed_manifest_tree)
    assert mpath == expected



#!/usr/bin/env python
# Ultra-optimized Manifest writing.
# (c) 2017-2023 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import gzip
import hashlib
import os
import os.path
import sys


def get_manifest_entry(t, path, relpath):
    sha512 = hashlib.sha512()
    blake2 = hashlib.blake2b()

    with open(path, 'rb') as f:
        buf = f.read()
        sha512.update(buf)
        blake2.update(buf)
        size = len(buf)

    return ('{} {} {} BLAKE2B {} SHA512 {}'.format(t, relpath,
            size, blake2.hexdigest(), sha512.hexdigest())).encode('utf8')


def generate_manifest_entries(out, topdir):
    compat_mode = False

    for dirpath, dirs, files in os.walk(topdir):
        if dirpath != topdir:
            for f in files:
                if f in ('Manifest', 'Manifest.gz'):
                    fp = os.path.join(dirpath, f)
                    out.append(get_manifest_entry('MANIFEST',
                            fp, os.path.relpath(fp, topdir)))
                    # do not descend
                    del dirs[:]
                    skip = True
                    break
            else:
                skip = False
            if skip:
                continue
        else:
            # enable compat mode for ebuild directories
            if any(f.endswith('.ebuild') and f != 'skel.ebuild' for f in files):
                compat_mode = True

        # skip dot-dirs
        dotdirs = [d for d in dirs if d.startswith('.')]
        for d in dotdirs:
            dirs.remove(d)

        for f in files:
            if f.startswith('Manifest') or f.startswith('.'):
                continue
            fp = os.path.join(dirpath, f)
            ep = os.path.relpath(fp, topdir)
            ftype = 'DATA'
            if compat_mode:
                if f.endswith('.ebuild') and f != 'skel.ebuild':
                    ftype = 'EBUILD'
                elif f == 'metadata.xml':
                    ftype = 'MISC'
                elif ep.startswith('files/'):
                    ftype = 'AUX'
                    ep = ep[6:]
            else:
                if f in ('timestamp', 'timestamp.chk', 'timestamp.commit',
                        'timestamp.x'):
                    continue

            out.append(get_manifest_entry(ftype, fp, ep))

    return compat_mode


def gen_manifest(top_dir):
    manifest_entries = []

    # load DIST and IGNORE entries from existing Manifest
    try:
        with open(os.path.join(top_dir, 'Manifest'), 'rb') as f:
            for l in f:
                if l.startswith(b'DIST') or l.startswith(b'IGNORE'):
                    manifest_entries.append(l.rstrip())
        had_manifest = True
    except FileNotFoundError:
        had_manifest = False

    # generate local file entries
    compat_mode = generate_manifest_entries(manifest_entries, top_dir)
    manifest_entries.sort()

    manifest_data = b'\n'.join(manifest_entries) + b'\n'
    if not compat_mode:
        with open(os.path.join(top_dir, 'Manifest.gz'), 'wb') as f:
            with gzip.GzipFile(fileobj=f, mode='wb', filename='', mtime=0) as gzf:
                gzf.write(manifest_data)
        if had_manifest:
            os.unlink(os.path.join(top_dir, 'Manifest'))
    else:
        with open(os.path.join(top_dir, 'Manifest'), 'wb') as f:
            f.write(manifest_data)


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print(f'Usage: {sys.argv[0]} <directory>...')
        sys.exit(1)

    for path in sys.argv[1:]:
        gen_manifest(path)



# gemato: Utility functions
# (c) 2017-2022 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later


class MultiprocessingPoolWrapper:
    """
    A portability wrapper for multiprocessing.Pool that supports
    context manager API (and any future hacks we might need).

    Note: the multiprocessing behavior has been temporarily removed
    due to unresolved deadlocks. It will be restored once the cause
    of the issues is found and fixed or worked around.
    """

    __slots__ = []

    def __init__(self, processes):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_cb):
        pass

    def map(self, func, it, chunksize=None):
        return map(func, it)

    def imap_unordered(self, *args, **kwargs):
        """
        Use imap_unordered() if available and safe to use. Fall back
        to regular map() otherwise.
        """
        return self.map(*args, **kwargs)


def path_starts_with(path, prefix):
    """
    Returns True if the specified @path starts with the @prefix,
    performing component-wide comparison. Otherwise returns False.
    """
    return prefix == "" or (path + "/").startswith(prefix.rstrip("/") + "/")


def path_inside_dir(path, directory):
    """
    Returns True if the specified @path is inside @directory,
    performing component-wide comparison. Otherwise returns False.
    """
    return ((directory == "" and path != "")
            or path.rstrip("/").startswith(directory.rstrip("/") + "/"))


def throw_exception(e):
    """
    Raise the given exception. Needed for onerror= argument
    to os.walk(). Useful for other callbacks.
    """
    raise e



# gemato: Profile behavior tests
# (c) 2017-2022 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import itertools
import os
import os.path

import pytest

import gemato.cli
from gemato.profile import (
    EbuildRepositoryProfile,
    BackwardsCompatEbuildRepositoryProfile,
    )
from gemato.recursiveloader import ManifestRecursiveLoader


DIRS = [
    'dev-foo',
    'dev-foo/bar',
    'dev-foo/bar/files',
    'eclass',
    'eclass/tests',
    'licenses',
    'metadata',
    'metadata/dtd',
    'metadata/glsa',
    'metadata/install-qa-check.d',
    'metadata/md5-cache',
    'metadata/md5-cache/dev-foo',
    'metadata/news',
    'metadata/news/2020-01-01-foo',
    'metadata/xml-schema',
    'profiles',
    'profiles/arch',
    'profiles/arch/foo',
    'profiles/desc',
    'profiles/updates',
]
PACKAGE_MANIFESTS = ['dev-foo/bar/Manifest']
EXPECTED_MANIFESTS = [
    'dev-foo/Manifest',
    'eclass/Manifest',
    'licenses/Manifest',
    'metadata/Manifest',
    'metadata/dtd/Manifest',
    'metadata/glsa/Manifest',
    'metadata/md5-cache/Manifest',
    'metadata/md5-cache/dev-foo/Manifest',
    'metadata/news/Manifest',
    'metadata/xml-schema/Manifest',
    'profiles/Manifest',
]
EXPECTED_TYPES = {
    EbuildRepositoryProfile: {
        'header.txt': 'DATA',
        'skel.ebuild': 'DATA',
        'skel.metadata.xml': 'DATA',
        'dev-foo/metadata.xml': 'DATA',
        'dev-foo/bar/bar-1.ebuild': 'DATA',
        'dev-foo/bar/metadata.xml': 'DATA',
        'dev-foo/bar/files/test.patch': 'DATA',
        'eclass/foo.eclass': 'DATA',
        'eclass/tests/foo.sh': 'DATA',
        'licenses/foo': 'DATA',
        'metadata/layout.conf': 'DATA',
        'metadata/projects.xml': 'DATA',
        'metadata/pkg_desc_index': 'DATA',
        'metadata/dtd/foo.dtd': 'DATA',
        'metadata/glsa/glsa-202001-01.xml': 'DATA',
        'metadata/install-qa-check.d/50foo': 'DATA',
        'metadata/md5-cache/dev-foo/bar-1': 'DATA',
        'metadata/news/2020-01-01-foo/2020-01-01-foo.en.txt': 'DATA',
        'metadata/news/2020-01-01-foo/2020-01-01-foo.en.txt.asc': 'DATA',
        'metadata/xml-schema/foo.xsd': 'DATA',
        'profiles/arch.desc': 'DATA',
        'profiles/categories': 'DATA',
        'profiles/eapi': 'DATA',
        'profiles/info_pkgs': 'DATA',
        'profiles/info_vars': 'DATA',
        'profiles/license_groups': 'DATA',
        'profiles/package.mask': 'DATA',
        'profiles/profiles.desc': 'DATA',
        'profiles/repo_name': 'DATA',
        'profiles/thirdpartymirrors': 'DATA',
        'profiles/use.desc': 'DATA',
        'profiles/use.local.desc': 'DATA',
        'profiles/arch/foo/eapi': 'DATA',
        'profiles/arch/foo/parent': 'DATA',
        'profiles/desc/foo.desc': 'DATA',
        'profiles/updates/1Q-2020': 'DATA',
    },
}
FILES = list(EXPECTED_TYPES[EbuildRepositoryProfile]) + [
    'metadata/timestamp',
    'metadata/timestamp.chk',
    'metadata/timestamp.commit',
    'metadata/timestamp.x',
]
EXPECTED_IGNORE = [
    'distfiles',
    'local',
    'lost+found',
    'packages',
    'metadata/timestamp',
    'metadata/timestamp.chk',
    'metadata/timestamp.commit',
    'metadata/timestamp.x',
    'metadata/dtd/timestamp.chk',
    'metadata/dtd/timestamp.commit',
    'metadata/glsa/timestamp.chk',
    'metadata/glsa/timestamp.commit',
    'metadata/news/timestamp.chk',
    'metadata/news/timestamp.commit',
    'metadata/xml-schema/timestamp.chk',
    'metadata/xml-schema/timestamp.commit',
]

EXPECTED_TYPES[BackwardsCompatEbuildRepositoryProfile] = (
    dict(EXPECTED_TYPES[EbuildRepositoryProfile]))
EXPECTED_TYPES[BackwardsCompatEbuildRepositoryProfile].update({
    'dev-foo/bar/bar-1.ebuild': 'EBUILD',
    'dev-foo/bar/metadata.xml': 'MISC',
    'dev-foo/bar/files/test.patch': 'AUX',
})

PKG_MANIFEST_SUFFIX = {
    EbuildRepositoryProfile: '.gz',
    BackwardsCompatEbuildRepositoryProfile: '',
}

ALL_PROFILES = [
    EbuildRepositoryProfile,
    BackwardsCompatEbuildRepositoryProfile,
]


@pytest.fixture
def test_repo(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp('profile-')
    for d in DIRS:
        os.mkdir(tmp_path / d)
    for f in FILES:
        with open(tmp_path / f, 'w'):
            pass
    yield tmp_path


@pytest.mark.parametrize(
    'profile,path,expected',
    [(profile, path, EXPECTED_TYPES[profile][path])
     for profile in ALL_PROFILES
     for path in EXPECTED_TYPES[profile]])
def test_get_entry_type_for_path(profile, path, expected):
    assert profile().get_entry_type_for_path(path) == expected


def make_entry_match(manifest_loader, path):
    entry = manifest_loader.find_path_entry(path)
    if entry is not None:
        if entry.tag == 'MANIFEST':
            return (entry.tag,
                    os.path.exists(
                        os.path.join(manifest_loader.root_directory,
                                     path)))
        elif entry.tag == 'AUX':
            return (entry.tag, entry.path, entry.aux_path)
        return (entry.tag,)


def make_expect_match(path, entry_type, manifests_exist=True):
    if entry_type == 'MANIFEST':
        return (entry_type, manifests_exist)
    elif entry_type == 'AUX':
        return (entry_type,
                os.path.join('files', os.path.basename(path)),
                os.path.basename(path))
    return (entry_type,)


@pytest.mark.parametrize('profile', ALL_PROFILES)
def test_update_entries_for_directory(test_repo, profile):
    m = ManifestRecursiveLoader(
        test_repo / 'Manifest',
        hashes=['SHA256', 'SHA512'],
        allow_create=True,
        profile=profile())
    m.update_entries_for_directory('')

    expected = dict(
        [(path, make_expect_match(path, entry_type))
         for path, entry_type in EXPECTED_TYPES[profile].items()] +
        [(path, make_expect_match(path, 'MANIFEST', manifests_exist=False))
         for path in EXPECTED_MANIFESTS + PACKAGE_MANIFESTS] +
        [(path, make_expect_match(path, 'IGNORE'))
         for path in EXPECTED_IGNORE])
    found = {path: make_entry_match(m, path) for path in expected}
    assert found == expected

    m.save_manifests()
    manifests_expected = dict(
        itertools.chain(
            ((path + '.gz', make_expect_match(path, 'MANIFEST'))
             for path in EXPECTED_MANIFESTS),
            ((path + PKG_MANIFEST_SUFFIX[profile],
             make_expect_match(path, 'MANIFEST'))
             for path in PACKAGE_MANIFESTS)))
    manifests_found = {path: make_entry_match(m, path)
                       for path in manifests_expected}
    assert manifests_found == manifests_expected

    m.assert_directory_verifies('')


@pytest.mark.parametrize('profile', ALL_PROFILES)
def test_cli_update(test_repo, profile):
    assert gemato.cli.main(['gemato', 'create',
                            '--profile', profile.name,
                            str(test_repo)]) == 0

    m = ManifestRecursiveLoader(test_repo / 'Manifest')
    expected = dict(
        [(path, make_expect_match(path, entry_type))
         for path, entry_type in EXPECTED_TYPES[profile].items()] +
        [(path + '.gz', make_expect_match(path, 'MANIFEST'))
         for path in EXPECTED_MANIFESTS] +
        [(path + PKG_MANIFEST_SUFFIX[profile],
          make_expect_match(path, 'MANIFEST'))
         for path in PACKAGE_MANIFESTS] +
        [(path, make_expect_match(path, 'IGNORE'))
         for path in EXPECTED_IGNORE])
    found = {path: make_entry_match(m, path) for path in expected}
    assert found == expected

    assert gemato.cli.main(['gemato', 'verify', str(test_repo)]) == 0


@pytest.mark.parametrize('profile', ALL_PROFILES)
def test_set_loader_options(test_repo, profile):
    m = ManifestRecursiveLoader(
        test_repo / 'Manifest',
        profile=profile(),
        allow_create=True)
    assert m.hashes is not None
    assert m.sort
    assert m.compress_watermark is not None
    assert m.compress_format is not None


@pytest.mark.parametrize('profile', ALL_PROFILES)
def test_regression_top_level_ignore_in_all_manifests(test_repo, profile):
    """Regression test for IGNORE wrongly applying to all Manifests"""
    m = ManifestRecursiveLoader(
        test_repo / 'Manifest',
        hashes=['SHA256', 'SHA512'],
        allow_create=True,
        profile=profile())
    m.update_entries_for_directory('')

    expected = {
        'distfiles': ('IGNORE',),
        'dev-foo/Manifest': ('MANIFEST', False),
        'dev-foo/distfiles': None,
    }
    found = {path: make_entry_match(m, path) for path in expected}
    assert found == expected
